/*
 * $Id: pcapdump.c,v 1.1.1.1 2002/02/23 02:47:02 kos Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#if HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "pcap.h"
#include "llog.h"

#define DEF_LINELEN 1024
int interrupted = 0;
extern char pcap_version[];

void usage(char *argv0, int code) {
  fprintf(stderr, "Usage: %s <options>\n", argv0);
  fprintf(stderr, "\t-i <interface name>\n");
  fprintf(stderr, "\t-v <verbose level>\n");
  fprintf(stderr, "\t-f <pcap filter expression\n");
  exit(code);
}

void
signal_handler(int sig) {
  switch(sig){
  case SIGINT:
    interrupted = 1;
  }
}

int
do_pcap_read(char *device, int snaplen, char *filter) {
  struct pcap_pkthdr pkthdr;
  unsigned char *pkt;
  char ebuf[PCAP_ERRBUF_SIZE];
  pcap_t *pd;
  struct bpf_program fcode;
  unsigned int if_net, if_mask;
  fd_set fds;
  struct timeval tm;
  int pcap_fd;
  unsigned int count;
  
  count = 0;

  if(!(pd = pcap_open_live(device, snaplen, 1, 100, ebuf))) {
    llog(0, "pcap_open_live(%s): %s\n", device, ebuf);
    return 2;
  }

  pcap_lookupnet(device, &if_net, &if_mask, ebuf);
  pcap_compile(pd, &fcode, filter, 1000, if_mask);
  pcap_setfilter(pd, &fcode);
  pcap_fd = pcap_fileno(pd);  
  
  llog(1, "pcap_datalink(%s) = %d\n", device, pcap_datalink(pd));

  while(1) {
    if(interrupted)
      break;
    
    FD_ZERO(&fds);
    FD_SET(pcap_fd, &fds);
    tm.tv_sec = 0;
    tm.tv_usec = 500;
    if(select(pcap_fd+1, &fds, NULL, NULL, NULL) < 0)
      break;
    
    if(!FD_ISSET(pcap_fd, &fds)) {
      continue;
    }
  
    if(!(pkt = (unsigned char *)pcap_next(pd, &pkthdr))) {
      llog(0, "ssid_sniff: pcap_next: %s\n", pcap_geterr(pd));
      break;
    }
    
    if(pkthdr.caplen != pkthdr.len) {
      llog(2, "pcapdump: warning: truncated packet\n");
    }
    
    llog(1, "> frame %d, length %u (0x%x), timestamp %u:%u\n",
	 count, pkthdr.len, pkthdr.len, pkthdr.ts.tv_sec, pkthdr.ts.tv_usec);
    llog_hexdump(1, pkt, pkthdr.len);
    count++;
  }

  pcap_freecode(&fcode);
  pcap_close(pd);
  return 0;
}

int
main(int argc, char **argv) {
  char c;
  int snaplen;
  char device[DEF_LINELEN];
  int verbose;
  char *pdev;
  char filter[DEF_LINELEN];

  snaplen = 2500;
  verbose = 1;
  device[0] = '\0';
  filter[0] = '\0';

  while((c = getopt(argc, argv, "i:v:f:")) != EOF) {
    switch(c) {
    case 'i':
      strncpy(device, optarg, DEF_LINELEN);
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    case 'f':
      strncpy(filter, optarg, DEF_LINELEN);
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  llog_init(LLOG_STDERR);
  llog_level(verbose);
  signal(SIGINT, signal_handler);
  llog(1, "%s: libpcap version %s\n", argv[0], pcap_version);

  if(!*device) {
    if(!(pdev = pcap_lookupdev(device))) {
      llog(0, "%s: No suitable device found\n", argv[0]);
      exit(1);
    } else {
      strncpy(device, pdev, PCAP_ERRBUF_SIZE);
      llog(1, "%s: Detected device %s\n", argv[0], device);
    }
  }
  
  do_pcap_read(device, snaplen, filter);

  llog_close();
  exit(0);
}

