/*
 * $Id: pcapset.c,v 1.1.1.1 2002/02/23 02:47:02 kos Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include "pcap.h"
#include "cli.h"
#include "llog.h"
#include "pcapset.h"

#define DEF_LINELEN 1024
#define TCPDUMP_MAGIC 0xa1b2c3d4

static pcap_linktype_t pcap_linktype[] = {
  {"LINKTYPE_NULL", DLT_NULL},
  {"LINKTYPE_ETHERNET", DLT_EN10MB},
  {"LINKTYPE_EXP_ETHERNET", DLT_EN3MB},
  {"LINKTYPE_AX25", DLT_AX25},
  {"LINKTYPE_PRONET", DLT_PRONET},
  {"LINKTYPE_CHAOS", DLT_CHAOS},
  {"LINKTYPE_TOKEN_RING", DLT_IEEE802},
  {"LINKTYPE_ARCNET", DLT_ARCNET},
  {"LINKTYPE_SLIP", DLT_SLIP},
  {"LINKTYPE_PPP", DLT_PPP},
  {"LINKTYPE_FDDI", DLT_FDDI},
  {"LINKTYPE_PPP_HDLC", 50},
  {"LINKTYPE_PPP_ETHER", 51},
  {"LINKTYPE_ATM_RFC1483", 100},
  {"LINKTYPE_RAW", 101},
  {"LINKTYPE_SLIP_BSDOS", 102},
  {"LINKTYPE_PPP_BSDOS", 103},
  {"LINKTYPE_C_HDLC", 104},
  {"LINKTYPE_ATM_CLIP", 106},
  {"LINKTYPE_IEEE802_11", 105},
  {"LINKTYPE_FR", 107},
  {"LINKTYPE_LOOP", 108},
  {"LINKTYPE_ENC", 109},
  {"LINKTYPE_LANE8023", 110},
  {"LINKTYPE_HIPPI", 111},
  {"LINKTYPE_HDLC", 112},
  {"LINKTYPE_LINUX_SLL", 113},
  {"LINKTYPE_LTALK", 114},
  {NULL, 0}
};

FILE *fp;
struct pcap_file_header hdr;
char file[DEF_LINELEN];

char *
linktype_i2a(unsigned int type) {
  int i;
  
  for(i=0; pcap_linktype[i].name; i++) {
    if(pcap_linktype[i].type == type)
      return pcap_linktype[i].name;
  }

  return "Unknown";
}

unsigned int
linktype_a2i(char *name) {
  int i;

  for(i=0; pcap_linktype[i].name; i++) {
    if(!strcmp(pcap_linktype[i].name, name)) {
      return pcap_linktype[i].type;
    }
  }
  
  return -1;
}

int
pcapdump_linktype(char *in, int type) {
  
  if(!(fp = fopen(in, "r+"))) {
    perror("fopen");
    return 1;
  }
  
  if(fread((void *)&hdr, sizeof(struct pcap_file_header), 1, fp) != 1) {
    perror("fread");
    fclose(fp);
    return 2;
  }

  if(hdr.magic != TCPDUMP_MAGIC) {
    fprintf(stderr, "wrong magic number: %x %x\n", hdr.magic, TCPDUMP_MAGIC);
    fclose(fp);
    return 3;
  }
  
  fprintf(stderr, "pcap_linktype: old %d new %d\n", hdr.linktype, type);
  hdr.linktype = type;

  fseek(fp, 0L, SEEK_SET);
  fwrite((void *)&hdr, sizeof(struct pcap_file_header), 1, fp);
  fclose(fp);
  return 0;
}

int
cmd_open(void *p, int argc, char **argv) {
  
  if(fp) {
    llog(0, "open: Seems like a file is open already: %s\n", file);
    return 0;
  }

  if(!(fp = fopen(argv[1], "r+"))) {
    llog(0, "fopen(%s): %s\n", argv[1], strerror(errno));
    return 0;
  }
  
  if(fread((void *)&hdr, sizeof(struct pcap_file_header), 1, fp) != 1) {
    llog(0, "fread: %s", strerror(errno));
    fclose(fp);
    fp = (FILE *)NULL;
    return 2;
  }

  if(hdr.magic != TCPDUMP_MAGIC) {
    fprintf(stderr, "wrong magic number: %x %x\n", hdr.magic, TCPDUMP_MAGIC);
    fclose(fp);
    fp = (FILE *)NULL;
    return 3;
  }
  
  llog(1, "ok\n");
  return 0;
}

int
cmd_quit(void *p, int argc, char **argv) {

  if(fp)
    fclose(fp);
  return 1;
}

int
cmd_info(void *p, int argc, char **argv) {

  if(!fp)
    return 0;

  llog(1, "pcap file: %s\n", file);
  llog(1, " version: %x:%x, timezone: %d\n", hdr.version_major, hdr.version_minor, hdr.thiszone);
  llog(1, " snaplen: %d, linktype %s (%d)\n", hdr.snaplen, linktype_i2a(hdr.linktype), hdr.linktype);
  return 0;
}

int
cmd_set(void *p, int argc, char **argv) {
  
  if(!strcmp(argv[1], "linktype")) {
    if((hdr.linktype = atoi(argv[2])))
      ;
    else if((hdr.linktype = linktype_a2i(argv[2])) == -1) {
      llog(0, "no such linktype\n");
      return 0;
    }

    fseek(fp, 0L, SEEK_SET);
    if(fwrite((void *)&hdr, sizeof(struct pcap_file_header), 1, fp) <=0)
      llog(0, "fwrite: %s\n", strerror(errno));
    else 
      llog(1, "set: linktype now %s (%d)\n", linktype_i2a(hdr.linktype), hdr.linktype);
    fflush(fp);
  } else
    llog(1, "dunno how to do that\n");
  return 0;
}

int
cmd_linktypes(void *p, int argc, char **argv) {
  int i;

  for(i=0; pcap_linktype[i].name; i++)
    llog(1, "%s = %d\n", pcap_linktype[i].name, pcap_linktype[i].type);
  return 0;
}

void
usage(char *argv0, int code) {
  fprintf(stderr, "Usage: %s -v <verbosity>\n", argv0);
  exit(code);
}

int
main(int argc, char **argv) {
  char c;
  int linktype = 1;
  cli_t *cli;
  int verbose;
  
  linktype = 1;
  verbose = 1;

  while((c = getopt(argc, argv, "v:")) != EOF) {
    switch(c) {
    case 'v':
      verbose = atoi(optarg);
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  fp = (FILE *)NULL;
  
  llog_init(LLOG_STDERR);
  llog_level(verbose);
  
  cli = cli_init(NULL);
  cli_set(cli, CLI_PROMPT, "pcapset> ");
  cli_register(cli, CLI_PARSE, (cmd_proc)cmd_open, "open", 2, 2, "Opens a pcap file");
  cli_register(cli, CLI_PARSE, (cmd_proc)cmd_quit, "quit", 0, 0, "duh");
  cli_register(cli, CLI_PARSE, (cmd_proc)cmd_info, "info", 0, 0, "info");
  cli_register(cli, CLI_PARSE, (cmd_proc)cmd_set,  "set",  3, 0, "sets");
  cli_register(cli, CLI_PARSE, (cmd_proc)cmd_linktypes, "linktypes", 0, 0, "lists all known linktypes");

  while(cli_main(cli) != 1)
    ;

  cli_destroy(cli);
  llog_close();
  exit(0);
}

