/*
 * $Id: ssidsniff.c,v 1.4 2002/03/03 22:07:47 kos Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#if HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "llog.h"
#include "hash.h"
#include "wifivendors.h"
#include "xpaudio.h"
#include "ssidsniff.h"

#ifdef HAVE_PCAP_H
#include "pcap.h"
#endif

#ifdef HAVE_NET_BPF_H
#include "net/bpf.h"
#endif

static char rotseq[] = "/-\\|";
int interrupted = 0;
extern char pcap_version[];


RETSIGTYPE
signal_handler(int sig) {
  switch(sig){
  case SIGINT:
    interrupted = 1;
  }
}

unsigned char *
mac2ascii(unsigned char *p) {
  static char out[DEF_LINELEN];
  
  snprintf(out, DEF_LINELEN, "%02x:%02x:%02x:%02x:%02x:%02x", 
	   p[0], p[1], p[2], p[3], p[4], p[5]);
  return out;
}

unsigned char *
hex2ascii(unsigned char *p, int l) {
  int i;
  static char out[DEF_LINELEN];
  char temp[DEF_LINELEN];

  out[0] = '\0';
  for(i=0; i<l; i++) {
    snprintf(temp, DEF_LINELEN, "%02x", p[i]);
    strncat(out, temp, DEF_LINELEN);
  }
  return out;
}

int
snarf_proc_datum(char *line, char *prefix, char *res, int len) {
  if(!strncmp(line, prefix, strlen(prefix))) {
    line[strlen(line)-1] = '\0';
    strncpy(res, line+strlen(prefix), len);
    return 0;
  }
  return 1;
}

int
packetinfo_zap(packetinfo_t *res) {

  memset((void *)res, '\0', sizeof(packetinfo_t));
  return 0;
}

unsigned char *
string_printable(unsigned char *in) {
  static char out[DEF_LINELEN];
  char octal[DEF_LINELEN];
  char *ret;

  memset(out, '\0', DEF_LINELEN);
  ret = out;

  while(*in) {
    if(*in >= 32 && *in <= 127) {
      *ret++ = *in;
    } else if(*in == '\\') {
      *ret++ = '\\';
      *ret++ = '\\';
    } else if(*in == '\a') {
      *ret++ = '\\';
      *ret++ = 'a';
    } else if(*in == '\b') {
      *ret++ = '\\';
      *ret++ = 'b';
    } else if(*in == '\n') {
      *ret++ = '\\';
      *ret++ = 'n';
    } else if(*in == '\r') {
      *ret++ = '\\';
      *ret++ = 'r';
    } else {
      snprintf(octal, DEF_LINELEN, "\\%03o", *in);
      strncat(ret, octal, strlen(octal));
      ret += strlen(octal);
    }
    in++;
  }
  return out;
}


int
dissect_80211(ssidsniff_t *s, unsigned char *p, unsigned int l, packetinfo_t *res) {
  f80211_mac_t *mac;
  frame_control_t *fc;
  beacon_header_t *bh;
  unsigned int le;
  unsigned int tlen, ttype;
  
  if(l<sizeof(f80211_mac_t)) {
    llog(6, "dissect: packet too small to have 802.11 MAC header\n");
    return 0;
  }
  
  mac = (f80211_mac_t *)p;  
  fc = (frame_control_t *)mac->fc;
  if(fc->version) {
    llog(6, "dissect: unhandled version number 0x%x\n", fc->version);
    return 0;
  }

  llog(4, "MAC 80211\n FC: version %u, type %u, subtype %u\n", 
       fc->version, fc->type, fc->subtype);

  res->timestamp = time(NULL);
  res->type = fc->type;
  res->subtype = fc->subtype;
  res->wepused = fc->wep;
  strncpy(res->mac, mac2ascii(mac->address2), DEF_LINELEN);
  strncpy(res->bss, mac2ascii(mac->address3), DEF_LINELEN);
  strncpy(res->dest, mac2ascii(mac->address1), DEF_LINELEN);
  
  switch(fc->subtype) {
  case MGT_PROBE_REQ:
    llog(4, "probe packet\n");
    break;
  case MGT_BEACON:
    if(sizeof(beacon_header_t)+sizeof(f80211_mac_t) > l) {
      llog(4, "dissect: bad frame size\n");
      return 0;
    }

    bh = (beacon_header_t *)(p+sizeof(f80211_mac_t));
    llog(4, " Type: BEACON\n");
    llog(4, " Destination: %s\n", mac2ascii(mac->address1));
    llog(4, " Source: %s\n", mac2ascii(mac->address2));
    llog(4, " BSS: %s\n", mac2ascii(mac->address3));
    llog(4, " Sequence: %02x%02x\n", mac->seq[1], mac->seq[0]);
    llog(4, "  beacon timestamp: %s\n", hex2ascii(bh->timestamp, 8));
    llog(4, "  beacon interval: %s\n", hex2ascii(bh->interval, 2));
    llog(4, "  beacon capability: %s\n", hex2ascii(bh->capability, 2));
    
    /* fixed parameters */
    res->ap  = bh->capability[0] & IEEE80211_CAPINFO_ESS;
    res->wepcapable = bh->capability[0] & IEEE80211_CAPINFO_PRIVACY;
    
    /* tagged parameters */
    for(le=sizeof(f80211_mac_t)+sizeof(beacon_header_t); le <= l; ) {
      
      if(le+1 > l)
	break;

      ttype=p[le];
      tlen=p[le+1];

      if(le+1+tlen > l) {
	llog(6, "dissect: problem decoding tag, ignoring\n");
	break;
      }

      switch(ttype) {
      case 0:
	llog(4, "   tag SSID: ");
	memset(res->ssid, '\0', DEF_LINELEN);
	memcpy(res->ssid, &p[le+2], tlen);
	llog(4, "%s\n", res->ssid);
	break;
      case 1:
	llog(4, "   tag RATES: \n");
	break;
      case 3:
	llog(4, "   tag DSPARAMS: ");
	res->channel = p[le+2];
	llog(4, "channel %d\n", res->channel);
	break;
      case 6:
	llog(4, "   tag IBSSPARAMS:\n");
	break;
      default:
	llog(4, "   unhandled tag %u\n", ttype);
	break;
      }

      le += tlen + 2;
    }

    break;
  default:
    llog(6, "dissect: unhandled type packet %d\n", *p>>4);
    break;
  }  

  return 0;
}

int
list_results(hash_t *reshash) {
  packetinfo_t *sample;
  int i;
  unsigned int l;
  char lhs[DEF_LINELEN];
  
  for(i=0; i<hash_nelem(reshash); i++) {
    memset(lhs, '\0', DEF_LINELEN);
    hash_index(reshash, i, lhs, DEF_LINELEN);
    hash_get(reshash, lhs, (void **)&sample, &l);
    
    llog(1, "%d: %s", sample->serial, ctime(&sample->timestamp));
    llog(1, "   MAC: %s (%s)\n", sample->mac, wifi_get_vendor(sample->mac));
    llog(1, "   BSS: %s\n", sample->bss);
    llog(1, "   Destination: %s\n", sample->dest);
    llog(1, "   Channel: %d\n", sample->channel);
    llog(1, "   SSID: %s\n", string_printable(sample->ssid));
    llog(1, "   WEP capable: %s\n", sample->wepcapable?"Y":"N");
    llog(1, "   WEP used: %s\n", sample->wepused?"Y":"N");
    llog(1, "   Access Point: %s\n", sample->ap?"Y":"N");
    llog(1, "   Bitrate: %s\n", sample->bitrate);
    llog(1, "   Signal level: %s, quality %s, noise %s\n", sample->level, sample->quality, sample->noise);
  }
  return 0;
}

int
proc_read_stats(packetinfo_t *res, char *device) {
  unsigned char line[DEF_LINELEN];
  char temp[DEF_LINELEN];
  FILE *fp;

  /* XXX: WTF? Can't we just have an ioctl? This is filthy */
  strncpy(res->noise, "NA", DEF_LINELEN);
  strncpy(res->level, "NA", DEF_LINELEN);
  strncpy(res->quality, "NA", DEF_LINELEN);
  strncpy(res->bitrate, "NA", DEF_LINELEN);

  snprintf(temp, DEF_LINELEN, "/proc/driver/aironet/%s/Status", device);
  if(!(fp = fopen(temp, "r")))
    return 1;
  
  while((fgets(line, DEF_LINELEN, fp))) {
    snarf_proc_datum(line, "Signal Strength: ", res->level, DEF_LINELEN);
    snarf_proc_datum(line, "Signal Quality: ", res->quality, DEF_LINELEN);
    snarf_proc_datum(line, "BitRate: ", res->bitrate, DEF_LINELEN);
  }
  
  fclose(fp);
  return 0;
}

int
runlog_add(FILE *fp, packetinfo_t *sample, runlog_mode_t mode) {
  
  if(!fp)
    return 1;

  switch(mode) {
  case machine:
    /* 
       TimeStampMilis;SSID;Channel;BooleanWep(1/0);BooleanIsAccessPoint(1/0)
       SignalStrength(integer);Noise(integer);DestAddr;SourceAddr;BSSID(mac)
    */
    fprintf(fp, "%lu;%s;%u;%u;%u;%s;%s;%s;%s;%s\n", 
	    time(NULL),
	    string_printable(sample->ssid),
	    sample->channel,
	    sample->wepcapable,
	    sample->ap,
	    sample->level,
	    sample->noise,
	    sample->dest,
	    sample->mac,
	    sample->bss
	    );
    break;
  case simple:
  default:
    fprintf(fp, "%d: %s", sample->serial, ctime(&sample->timestamp));
    fprintf(fp, "   MAC: %s (%s)\n", sample->mac, wifi_get_vendor(sample->mac));
    fprintf(fp, "   BSS: %s\n", sample->bss);
    fprintf(fp, "   Destination: %s\n", sample->dest);
    fprintf(fp, "   Channel: %d\n", sample->channel);
    fprintf(fp, "   SSID: %s\n", string_printable(sample->ssid));
    fprintf(fp, "   Bitrate: %s\n", sample->bitrate);
    fprintf(fp, "   WEP capable: %s\n", sample->wepcapable?"Y":"N");
    fprintf(fp, "   WEP used: %s\n", sample->wepused?"Y":"N");
    fprintf(fp, "   Access Point: %s\n", sample->ap?"Y":"N");
    fprintf(fp, "   Signal level: %s, quality %s, noise %s\n", sample->level, sample->quality, sample->noise);
  }

  fflush(fp);
  return 0;
}

int
ssid_sniff(ssidsniff_t *s) {
  infosink_t count;
  struct pcap_pkthdr pkthdr;
  unsigned char *pkt;
  packetinfo_t sample;
  hash_t *reshash;
  fd_set fds;
  struct timeval tm;
  char junk[DEF_LINELEN];
  
  if(pcap_datalink(s->pd) != DLT_IEEE802_11) {
    llog(1, "ssid_sniff: datalink type %d isn't 802.11, continuing anyway\n", pcap_datalink(s->pd));
  }

  reshash = hash_new(0);
  count.data = count.beacon = count.total = count.martian = count.wep = 0;
  
  while(1) {
    if(interrupted || (s->maxcount && count.total >= s->maxcount))
      break;

    /* XXX: according to the pcap docs, this isn't portable */
    /* we're linux specific now anyway */
    if(s->mode == live) {
      FD_ZERO(&fds);
      FD_SET(s->pcap_fd, &fds);
      FD_SET(0, &fds);
      tm.tv_sec = 0;
      tm.tv_usec = 500;
      if(select(s->pcap_fd+1, &fds, NULL, NULL, NULL) < 0)
	break;

      if(FD_ISSET(0, &fds)) {
	read(0, junk, DEF_LINELEN);
	list_results(reshash);
	llog(1, "\n");
      }

      if(!FD_ISSET(s->pcap_fd, &fds))
	continue;
    }

    if(!(pkt = (unsigned char *)pcap_next(s->pd, &pkthdr))) {
      llog(0, "ssid_sniff: pcap_next: %s\n", pcap_geterr(s->pd));
      break;
    }

    /* When capturing raw, save all packets we see */
    if(s->pd_dump && !s->strip_80211) {
      pcap_dump((u_char *)s->pd_dump, &pkthdr, pkt);
    }

    if(pkthdr.caplen != pkthdr.len) {
      llog(7, "ssid_sniff: packet truncated, discarding\n");;
      continue;
    }

    count.total++;
    packetinfo_zap(&sample);
    dissect_80211(s, pkt, pkthdr.len, &sample);    
    
    llog_hexdump(3, pkt, pkthdr.len);
    
    if(sample.type == 0 && sample.subtype == 8) {     /* beacon packet */
      if(!hash_exists(reshash, sample.mac)) {
	llog(1, "ssid_sniff: found new beacon: %s (%s)\n", 
	     string_printable(sample.ssid), wifi_get_vendor(sample.mac));
	sample.serial = count.beacon;
	proc_read_stats(&sample, s->device);
	hash_add(reshash, sample.mac, &sample, sizeof(sample));
	runlog_add(s->runlog, &sample, s->runlog_mode);
	count.beacon++;
      }     
    } else if(sample.type == 2 && !sample.wepused) {   /* data packet */
      llog(2, "ssid_sniff: data packet, whooo\n");
      llog_hexdump(2, pkt, pkthdr.len);
      count.data++;
    } else if(sample.type == 2 && sample.wepused) {
      llog(2, "ssid_sniff: data packet, encrypted\n");
      count.wep++;      
    } else {                                          /* 802.11 martian */
      llog(3, "ssid_sniff: unknown type packet found: %u %u\n", sample.type, sample.subtype);
      llog_hexdump(3, pkt, pkthdr.len);
      count.martian++;
    }

    /* When capturing cooked ether style frames, only save data */
    /* XXX: pcap makes this really difficult. Investigating alternatives */
    if(s->pd_dump && s->strip_80211 && sample.type==2) {
      pkt += 0x20;
      pkthdr.len -= 0x20;
      pkthdr.caplen -= 0x20;
      pcap_dump((u_char *)s->pd_dump, &pkthdr, pkt);
    }
    
    if(s->mode_geiger) {
      if(s->xpa)
	xpaudio_send(s->xpa, pkt, pkthdr.len>DEF_GEIGER_L?DEF_GEIGER_L:pkthdr.len);
      else
	printf("\a");
    }

    //       --------------------------------------------------------------------------------
    llog(1, "%c %u total, %u beacons, %u plaintext, %u wep, %u martians                     \r", 
	 rotseq[(count.total)%4], count.total, count.beacon, count.data, count.wep, count.martian);
  }

  list_results(reshash);
  hash_destroy(reshash);
  return 0;
}

void
usage(char *argv0, int code) {
  fprintf(stderr, "Usage: %s <options>\n", argv0);
  fprintf(stderr, "\t-i <device> Set the device to listen on\n");
  fprintf(stderr, "\t-s <snaplen> pcap maximum snarfed length\n");
  fprintf(stderr, "\t-v <verbosity> The higher, the noisier\n");
  fprintf(stderr, "\t-f <filter> pcap filter to use\n");
  fprintf(stderr, "\t-c <maxcount> Set maximum packets to read, then exit\n");
  fprintf(stderr, "\t-g Geiger counter mode. Beep for every packet received.\n");
  fprintf(stderr, "\t-w <file> tcpdump capture file for everything received\n");
  fprintf(stderr, "\t-W   When capturing to file, only save 802.3 portion\n"); 
  fprintf(stderr, "\t-r <file> tcpdump capture file to read packets from\n");
  fprintf(stderr, "\t-l <runlog> Text file to keep findings\n");
  fprintf(stderr, "\t-L   When capturing to text file, use machine parseable format\n");
  fprintf(stderr, "\t-V version number\n");
  exit(code);
}

int
main(int argc, char **argv) {
  char c;
  int verbose;
  char *pdev;
  ssidsniff_t s;
  
  /* defaults */
  verbose = 1;
  s.maxcount = 0;
  s.device[0] = '\0';
  s.mode = live;
  s.mode_geiger = 0;
  s.strip_80211 = 0;
  s.pcap_filter[0] = '\0';
  s.pcap_snaplen = 750;
  s.pcap_datadump[0] = '\0';
  s.pcap_promiscuous = 1;
  s.interactive = isatty(0);
  s.xpa = (xpaudio_t *)NULL;
  s.runlog = (FILE *)NULL;
  s.runlog_mode = simple;

  while((c = getopt(argc, argv, "i:s:f:v:c:gw:r:WVl:L"))!=EOF) {
    switch(c) {
    case 'i':
      s.mode = live;
      strncpy(s.device, optarg, PCAP_ERRBUF_SIZE);
      break;
    case 's':
      s.pcap_snaplen = atoi(optarg);
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    case 'f':
      strncpy(s.pcap_filter, optarg, DEF_LINELEN);
      break;
    case 'c':
      s.maxcount = atoi(optarg);
      break;
    case 'g':
      s.mode_geiger = 1;
      break;
    case 'w':
      strncpy(s.pcap_datadump, optarg, DEF_LINELEN);
      break;
    case 'r':
      s.mode = file;
      strncpy(s.device, optarg, PCAP_ERRBUF_SIZE);
      break;
    case 'W':
      s.strip_80211 = 1;
      break;
    case 'V':
      printf("%s %s\n", PACKAGE, VERSION);
      exit(0);
    case 'l':
      if(!(s.runlog = fopen(optarg, "w")))
	fprintf(stderr, "%s: fopen(%s): %s\n", argv[0], optarg, strerror(errno));
      break;
    case 'L':
      s.runlog_mode = machine;
      fprintf(s.runlog, "TimeStampMilis;SSID;Channel;BooleanWep(1/0);BooleanIsAccessPoint(1/0);SignalStrength(integer);Noise(integer);DestAddr;SourceAddr;BSSID(mac)\n");
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  llog_init(LLOG_STDERR);
  llog_level(verbose);

  /* Initialize pcap part */
  if(s.mode == live && !*s.device) {
    if(!(pdev = pcap_lookupdev(s.device))) {
      llog(0, "%s: No suitable device found\n", argv[0]);
      exit(1);
    } else {
      strncpy(s.device, pdev, PCAP_ERRBUF_SIZE);
      llog(1, "%s: Detected device %s\n", argv[0], s.device);
    }
  }
  
  if(s.mode == live) {
    if(!(s.pd = pcap_open_live(s.device, s.pcap_snaplen, s.pcap_promiscuous, 100, s.pcap_ebuf))) {
      llog(0, "pcap_open_live(%s): %s\n", s.device, s.pcap_ebuf);
      return 2;
    }
    pcap_lookupnet(s.device, &s.pcap_if_net, &s.pcap_if_mask, s.pcap_ebuf);
    pcap_compile(s.pd, &s.pcap_fcode, s.pcap_filter, 1000, s.pcap_if_mask);
    pcap_setfilter(s.pd, &s.pcap_fcode);
    s.pcap_fd = pcap_fileno(s.pd);  
  } else {
    if(!(s.pd = pcap_open_offline(s.device, s.pcap_ebuf))) {
      llog(0, "pcap_open_offline(%s): %s\n", s.device, s.pcap_ebuf);
      return 3;
    }
  }

  s.pd_dump = (pcap_dumper_t *)NULL;
  if(s.pcap_datadump[0]) {
    if(!(s.pd_dump = pcap_dump_open(s.pd, s.pcap_datadump))) {
      llog(0, "pcap_open_offline(%s): %s\n", s.pcap_datadump, pcap_geterr(s.pd));
    }
  }
  llog(1, "%s: libpcap version %s\n", argv[0], pcap_version);
  
  /* want sound? */
  if(s.mode_geiger) {
    s.xpa = xpaudio_new(NULL);
    llog(1, "%s: geiger mode on: %s\n", argv[0], xpaudio_describe(s.xpa));
  }
  
  /* start it up */
  signal(SIGINT, signal_handler);
  llog(1, "%s: Starting sniffing with filter=%s on %s\n", argv[0], s.pcap_filter, s.device);
  ssid_sniff(&s);

  /* shutdown pcap */
  if(s.mode == live)
    pcap_freecode(&s.pcap_fcode);
  pcap_close(s.pd);

  if(s.pd_dump)
    pcap_dump_close(s.pd_dump);

  if(s.xpa)
    xpaudio_destroy(s.xpa);

  if(s.runlog)
    fclose(s.runlog);

  llog_close();
  exit(0);
}

